#include "stdio.h"

/* The Zen text editor. Inspired by UNIX ed. 
   "It's good ENOUGH, OK?"
 */

#define MAXLINES 1000
#define MAXLENGTH 1000
char *line[MAXLINES], buf[MAXLENGTH];
int numlines;

void clearbuf() {
  int i;
  for(i=0;i<MAXLINES;i++) if (line[i]) free(line[i]);
  numlines=0;
}

int getline(FILE *f, int l) {
  fgets(buf,MAXLENGTH,f);
  if (buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]='\0';
  if ((f==stdin)&&(!strcmp(buf,"."))) return 2;
  if (line[l]) free(line[l]);
  line[l]=malloc(strlen(buf)+1);
  if (line[l]) {
    strcpy(line[l],buf);
  } else {
    printf("insufficient memory\n");
    return NULL;
  }
  return 1;
}

int loadfile(char *name) {
  FILE *f;

  f=fopen(name,"r");
  if (!f) {
    printf("couldn't open '%s'\n",name);
    return 0;
  }
  printf("loading from %s\n",name);
  clearbuf();
  while(!feof(f)) {
    if (!getline(f,numlines)) break; else numlines++;
    if (numlines>MAXLINES) {
      numlines--;
      printf("too many lines -- file truncated\n");
      break;
    }
  }
  fclose(f);
  printf("loaded %d lines from '%s'\n",numlines,name);
  return numlines;
}

int savefile(char *name) {
  FILE *f;
  int i;

  f=fopen(name,"w");
  if (!f) {
    printf("couldn't open '%s'\n",name);
    return 0;
  }
  for(i=0;i<numlines;i++) fprintf(f,"%s\n",line[i]);
  fclose(f);
  printf("wrote %d lines to '%s'\n",numlines,name);
  return 1;
}

int main(int argc, char **argv) {
  int i, j, first, last, running;
  char *s, fn[MAXLENGTH];

  fn[0]='\0';
  numlines=0;
  for(i=0;i<MAXLINES;i++) line[i]=NULL;

  if (argc>1) {
    strcpy(fn,argv[1]);
    if (!loadfile(fn)) {
      printf("%s: new file\n",fn);
    }
  }
  running=1;
  while(running) {
    putchar('-');
    fgets(buf, MAXLENGTH, stdin);
    if (buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]='\0';
    first=last=0;
    if (first=atoi(buf)) {
      s=index(buf,',');
      if (s) last=atoi(s+1);
    }

    if (*buf=='!') {
      system(buf+1);
    } else {
      switch(buf[strlen(buf)-1]) {
      case 'l':
	if (first&&!last) {
	  last=first;
	} else {
	  if (!first) first=1;
	  if (!last) last=numlines;
	}
	for(i=first-1;i<last;i++) if (line[i]) printf("%d. %s\n",i+1,line[i]);
	break;
      case 'c':
	if (!first) {
	  printf("must specify line number\n");
	} else {
	  printf("%d. ", first);
	  getline(stdin, first-1);
	}
	break;
      case 'i':
	if (!first) {
	  printf("must specify line number\n");
	} else {
	  if (numlines>=MAXLINES) {
	    printf("no room to insert\n");
	  } else {
	    for(;;) {
	      s=line[first-1];
	      line[first-1]=NULL;
	      printf("%d. ",first);
	      if (2!=getline(stdin,first-1)) {
		for(i=numlines;i>first;i--) line[i]=line[i-1];
		line[first]=s;
		numlines++;
		first++;
	      } else {
		line[first-1]=s;
		break; /* out of the while */
	      }
	    }
	  }
	}
	break;
      case 'a':
	for(;;) {
	  printf("%d. ",numlines+1);
	  if (2!=getline(stdin, numlines)) {
	    numlines++;
	  } else {
	    break;
	  }
	}
	break;
      case 'd':
	if (!first) {
	  printf("must specify line number\n");
	} else {
	  if (!last) last=first;
	  for(j=first-1;j<last;j++) {
	    /* This seems wrong but is actually right. */
	    if (line[first-1]) free(line[first-1]);
	    for(i=first-1;i<numlines;i++) line[i]=line[i+1];
	    numlines--;
	  }
	}
	break;
      case 'n':
	clearbuf();
	break;
      case 'r':
	printf("filename: ");
	fgets(fn, MAXLENGTH, stdin);
	loadfile(fn);
	break;
      case 'q':
	running=0;
	/* fall through */
      case 'w':
	buf[strlen(buf)-1]='\0';
	if (*buf) strcpy(fn, buf);
	if (!*fn) {
	  printf("must specify a filename\n");
	  running=1;
	} else {
	  if (!savefile(fn)) running=1; /* don't quit if not saved */
	}
	break;
      case 'x':
	running=0;
	break;
      default:
	printf("unknown command\n");
	break;
      }
    }
  }
  
  clearbuf();
  return 0;
}
