/* 
   asple-rpn: a language. "Adam's Simple Programming Language Environment"?
   This is my take on Reverse Polish LISP, with on-the-fly GC.
   azz@josstix.demon.co.uk

FINISH "join"

*/

#include "stdio.h"

/* #define DEBUG */
#define BYTETYPE unsigned char
#define WORDTYPE unsigned int
#define CORESIZE 4000
#define NUMFBLOCKS 100
#define BUFSIZE 100

#define L(x) ((x)%256)
#define H(x) ((x)/256)
#define WORD(h,l) ((h)*256+(l))
#define ISSPACE(c) (((c)==' ') || ((c)=='\t') || ((c)=='\n') || ((c)=='\r'))

/* "items" -- cons cells in LISP -- look like this:
WORDTYPE next
BYTETYPE variety
WORDTYPE referent
WORDTYPE length          for blocks (strings and the like)
*/

#define ITEMSIZE 8
#define INEXTL(x) core[x]
#define INEXTH(x) core[x+1]
#define IVAR(x) core[x+2]
#define IREFL(x) core[x+3]
#define IREFH(x) core[x+4]
#define ILENL(x) core[x+5]
#define ILENH(x) core[x+6]
#define ICOUNT(x) core[x+7]

/* Valid referent values */
#define RNONE 0 /* no data */
#define RINT 1 /* integer */
#define RSTRING 2 /* string or text, length in len */
#define RLIST 3 /* list */
#define RCMD 4 /* command */

/* True and false */

#define NIL 0
#define T 1

/* size of free blocks, base of free blocks, number of free blocks */
WORDTYPE fbsize[NUMFBLOCKS], fbbase[NUMFBLOCKS], numfb;

/* core memory */
BYTETYPE core[CORESIZE];

#define STACKSIZE 400
WORDTYPE stack[STACKSIZE], sp;

/* buffer for string being read */
char buf[BUFSIZE];

void fbdump() {
  int y;

  printf("free blocks: ");
  for (y=0; y<numfb; y++) printf("%d[%d] ",fbbase[y],fbsize[y]);
  printf("\n");
}

/* allocate size bytes of memory */
WORDTYPE take(WORDTYPE size) {
  int i,j;
  WORDTYPE r;
  
  r=0;
  if (numfb) for (i=0; (i<numfb)&&(!r); i++) if (fbsize[i]>=size) {
    r=fbbase[i];
    fbbase[i]=fbbase[i]+size;
    fbsize[i]=fbsize[i]-size;
    if (!fbsize[i]) {
      j=i;
      while(j<=(numfb-1)) {
	fbbase[j]=fbbase[j+1];
	fbsize[j]=fbsize[j+1];
	j++;
      }
      numfb--;
    }
  }
#ifdef DEBUG
  printf("take at %d size %d\n", r, size);
  fbdump();
#endif
  return r;
}

/* return size bytes of memory at base */
void give(WORDTYPE base, WORDTYPE size) {
  int i,j;

#ifdef DEBUG
  printf("give at %d size %d\n", base, size);
#endif
  i=0;
  while ((fbbase[i]<base) && (i<numfb)) i++;
  if (base) if (i && ((fbbase[i-1]+fbsize[i-1])==base)) { /* join left? */
    fbsize[i-1] = fbsize[i-1] + size;
    if ((i<numfb) && ((fbbase[i-1]+fbsize[i-1])==fbbase[i])) { /* then right? */
      fbbase[i] = fbbase[i-1];
      fbsize[i] = fbsize[i] + fbsize[i-1];
      j=i;
      while(j<numfb) {
	fbbase[j-1] = fbbase[j];
	fbsize[j-1] = fbsize[j];
	j++;
      }
      numfb--;
    }
  } else {
    if ((i<numfb) && ((base+size)==fbbase[i])) { /* join right? */
      fbbase[i] = base;
      fbsize[i] = fbsize[i] + size;
    } else {
      j=numfb;
      while(j>i) {
	fbbase[j] = fbbase[j-1];
	fbsize[j] = fbsize[j-1];
	j--;
      }
      fbbase[i] = base;
      fbsize[i] = size;
      numfb++;      
    }
  }
#ifdef DEBUG
  fbdump();
#endif
}

void setnext(WORDTYPE addr, WORDTYPE next) {
  INEXTL(addr) = L(next);
  INEXTH(addr) = H(next);
}

WORDTYPE next(WORDTYPE addr) {
  if (!addr) printf("Asked to next null\n");
  return WORD(INEXTH(addr), INEXTL(addr));
}

void setref(WORDTYPE addr, WORDTYPE referent) {
  IREFL(addr) = L(referent);
  IREFH(addr) = H(referent);
}

WORDTYPE ref(WORDTYPE addr) {
  return WORD(IREFH(addr), IREFL(addr));
}

void setlen(WORDTYPE addr, WORDTYPE lenn) {
  ILENL(addr) = L(lenn);
  ILENH(addr) = H(lenn);
}

WORDTYPE len(WORDTYPE addr) {
  return WORD(ILENH(addr), ILENL(addr));
}

WORDTYPE newitem(BYTETYPE variety, WORDTYPE referent, WORDTYPE length) {
  WORDTYPE r;

  r=take(ITEMSIZE);
  IVAR(r) = variety;
  ICOUNT(r) = 1;
  setnext(r, NIL);
  setref(r, referent);
  setlen(r, length);

#ifdef DEBUG
  printf("new item v:%d r:%d l:%d at %d\n", variety, referent, length, r);
#endif
  return r;
}

WORDTYPE refer(WORDTYPE addr) {
  WORDTYPE t;

#ifdef DEBUG
  printf("Referring to %d (count before ref %d)\n", addr, ICOUNT(addr));
#endif DEBUG
  t=ref(addr);
  if ((IVAR(addr)==RLIST)&&t) refer(t);
  t=next(addr);
  if (t) refer(t);
  ICOUNT(addr)++;
  return addr;
}

void unrefer(WORDTYPE addr) {
  WORDTYPE t;
  
#ifdef DEBUG
  printf("unreferring from %d (count before unrefer %d)\n", addr, ICOUNT(addr));
#endif DEBUG
  t=ref(addr);
  if ((IVAR(addr)==RLIST)&&t) unrefer(t);
  t=next(addr);
  if (t) unrefer(t);
  if (!(--ICOUNT(addr))) {
    if (len(addr)) give(ref(addr), len(addr));
    give(addr, ITEMSIZE);
  }
}

int lenstr(char *s) {
  int i;

  i=0;
  while(*(s++)) i++;
  return i;
}

int samestr(char *s, char *t) {
  if (lenstr(s) != lenstr(t)) return NIL;
  while (*s) if (*(s++) != *(t++)) return NIL;
  return T;
}

WORDTYPE takestring(char *s) {
  WORDTYPE r, t;

  t=r=take(lenstr(s)+1);
#ifdef DEBUG
  printf("writing \"%s\" into core at %d\n", s, t);
#endif DEBUG
  while(*s) core[t++] = (WORDTYPE) *(s++);
  core[t] = NIL;
  return r;
}

void getword() {
  char c, *s;

  c=' ';
  while(ISSPACE(c)) c=getchar();
  s=buf;
  *s=c;
  while(!ISSPACE(c)) {
    c=getchar();
    if (!ISSPACE(c)) *(++s)=c;
  }
  *(++s)='\0';
}

/* This should be called after the list left sep has been read. */
WORDTYPE getlist() {
  WORDTYPE rf, prev, this, first, len;
  BYTETYPE t;
  int f;
  char *c;

  len=first=prev=NIL;
  f=T;
  while(f) {
    getword();
    f=!samestr(buf, ")");
    if (f) {
      t=RSTRING;
      len=NIL;
      if ((buf[0]>='0') && (buf[0] <='9')) {
	t=RINT;
	rf=0;
	c=buf;
	/* Who needs atoi? This does just as well. :-) */
	while(*c) rf=(rf*10)+(*(c++)-'0');
      } else if (samestr(buf,"(")) {
	t=RLIST;
	rf=getlist();
      } else if (buf[0]=='\'') {
	t=RSTRING;
	rf=takestring(buf+1);
	len=lenstr(buf+1)+1;
      } else {
	t=RCMD;
	rf=takestring(buf);
	len=lenstr(buf)+1;
      }
      this=newitem(t,rf,len);
      if (prev) setnext(prev,this);
      if (!first) first=this;
      prev=this;
    }
  }
  return first;
}

void coretobuf(WORDTYPE kk) {
  char *b;
  WORDTYPE k;

  k=kk;
  b=buf;
  while(core[k]) *(b++)=(char)core[k++];
  *b='\0';
}

void print(WORDTYPE thing) {
  WORDTYPE t;

#ifdef DEBUG
  printf("PRINT OUTPUT>> ");
#endif
  t=thing;
  while(t) {
    switch(IVAR(t)) {
    case RNONE:
      printf("[none] ");
      break;
    case RINT:
      printf("%d ", ref(t));
      break;
    case RCMD:
      coretobuf(ref(t));
      printf("%s ", buf);
      break;
    case RSTRING:
      coretobuf(ref(t));
      printf("\"%s\" ", buf);
      break;
    case RLIST:
      printf("( ");
      print(ref(t));
      printf(") ");
      break;
    default:
      printf("[???] ");
      break;
    }
    t=next(t);
  }
#ifdef DEBUG
  printf("<<\n");
#endif
}

WORDTYPE push(WORDTYPE l) {
#ifdef DEBUG
  printf("push %d (sp:%d)\n", l, sp);
#endif
  stack[sp++] = l;
  return l;
}

WORDTYPE pop() {
#ifdef DEBUG
  printf("pop %d (sp:%d)\n", stack[sp-1], sp);
#endif
  return stack[--sp];
}

/* copies a block */
WORDTYPE dupcore(WORDTYPE c, WORDTYPE l) {
  WORDTYPE r, ll, cc, rr;

  cc=c;
  rr=r=take(l);
  ll=l;
  while(ll--) core[rr++]=core[cc++];
  return r;
}

/* copies the head of a list */
WORDTYPE head(WORDTYPE t) {
  if (IVAR(t)==RLIST) refer(ref(t));
  if (len(t)) {
    return newitem(IVAR(t), dupcore(ref(t), len(t)), len(t));
  } else {
    return newitem(IVAR(t), ref(t), len(t));
  }
}

WORDTYPE numpop() {
  WORDTYPE a, b;
  
  a=pop();
  if (IVAR(a)==RINT) {
    b=ref(a);
  } else {
    printf("Not int\n");
    b=0;
  }
  unrefer(a);
#ifdef DEBUG
  printf("popped number %d\n", b);
#endif
  return b;
}

WORDTYPE numpush(WORDTYPE n) {
  WORDTYPE a;

  a=newitem(RINT, n, NIL);
  push(a);
#ifdef DEBUG
  printf("pushed number %d\n", a);
#endif
  return a;
}

/* interpret an asple-rpn list */
void interpret(WORDTYPE l) {
  WORDTYPE t, s, r;

  t=l;
  while(t) {
#ifdef DEBUG
    printf("interpret %d\n", t);
    fbdump();
#endif
    if (IVAR(t)==RCMD) {
      coretobuf(ref(t));
#ifdef DEBUG
      printf("... which is \"%s\"\n", buf);
#endif
      if (samestr(buf,"print")) {
	s=pop();
	print(s);
	unrefer(s);
      } else if (samestr(buf,"drop")) {
	unrefer(pop());
      } else if (samestr(buf,"dup")) {
	s=pop();
	push(s);
	push(head(s));
      } else if (samestr(buf,"swap")) {
	s=pop();
	r=pop();
	push(s);
	push(r);
      } else if (samestr(buf,"+")) {
	numpush(numpop()+numpop());
      } else if (samestr(buf,"-")) {
	s=numpop();
	numpush(numpop()-s);
      } else if (samestr(buf,"*")) {
	numpush(numpop()*numpop());
      } else if (samestr(buf,"/")) {
	s=numpop();
	numpush(numpop()/s);
      } else if (samestr(buf,"if")) {
	r=pop();
	s=numpop();
	if (s) interpret(r);
	unrefer(r);
      } else if (samestr(buf,"while")) {
	r=pop();
	do {
	  interpret(r);
	} while (numpop());
	unrefer(r);
      } else if (samestr(buf,"=")) {
	numpush(numpop()==numpop());
      } else if (samestr(buf,"<")) {
	r=numpop();
	numpush(numpop()<r);
      } else if (samestr(buf,">")) {
	r=numpop();
	numpush(numpop()>r);
      } else if (samestr(buf,"t")) {
	numpush(T);
      } else if (samestr(buf,"nil")) {
	numpush(NIL);
      } else if (samestr(buf,"head")) {
	r=pop();
	push(head(r));
	unrefer(r);
      } else if (samestr(buf,"tail")) {
	r=pop();
	push(next(r));
	unrefer(r);
	refer(next(r)); /* I think this is correct behaviour. */
      } else if (samestr(buf,"var")) {
	r=pop();
	numpush(IVAR(r));
	unrefer(r);
      } else if (samestr(buf,"join")) {
	/* This should copy the lists---create a new list from the original
	   parts. A copylist() function? This would work similar to head().
	*/
	s=pop();
	r=pop();
	while(next(r)) r=next(r);
	
      } else {
	/* user word */
      }
    } else {
#ifdef DEBUG
      printf("... which is an object\n");
#endif
      if (IVAR(t)==RLIST) {
	push(ref(t));
	refer(ref(t));
      } else {
	push(head(t));
      }
    }
    t=next(t);
  }
}

/* amt of free mem */
WORDTYPE freemem() {
  int i;
  WORDTYPE t;

  t=0;
  for(i=0;i<numfb;i++) t=t+fbsize[i];
  return t;
}

int main(int argc, char **argv) {
  WORDTYPE l;
  int seq;

  /* give all the memory, to put it in the table */
  give(1, CORESIZE-1);

  sp=0;

  /* go into the interpreter loop */

  seq=0;
  while(1) {
    printf("\n[M:%d S:%d C:%d] ( ", freemem(), sp, seq); fflush(stdout);
    seq++;
    l=getlist();
#ifdef DEBUG
    printf("shell: user command list is at %d, about to interpret\n", l);
#endif
    interpret(l);
#ifdef DEBUG
    printf("shell: about to unrefer user list at %d\n", l);
#endif
    unrefer(l);
  }

  /*exit(0);*/ return 0;
}

