/* zen, azz@josstix.demon.co.uk
 */

#define MEMORYSIZE (56*1024)
#define VERSION "0.001"

/* --- stuff that should be in zen.h --- */

#define NULL 0
#define FALSE 0
#define TRUE 1

/* types of FILE */
#define TYPE_DEV 0
#define TYPE_FILE 1
/* IDEA: Have a TYPE_STRING, which would make sprintf a darn sight
   easier to implement. */

/* devices */
#define DEV_CONSOLE 0
#define DEV_SERIAL 1
#define DEV_PARALLEL 2

struct filedescriptor {
  char type;
  char device;
  char unit;
  /* more stuff for FILEs... */
};

typedef struct filedescriptor FILE;
FILE console, *stdin, *stdout;
extern int ramstart;

/* K&R.. hmm.. */
typedef int ALIGN; /* OK, this isn't a PDP-11, but what the hell */
union header {
  struct {
    union header *ptr;
    unsigned size;
  } s;
  ALIGN x;
};
typedef union header HEADER;

/* --- protos --- */

void main();
int strcmp(char *p, char *q);
FILE *fopen(char *name, char *mode);
void fclose(FILE *file);
void init_os();
void fprintf(FILE *f, char *s, ...);
void printf(char *s, ...);
void doprintf(FILE *f, char *s, int *a);
void fputc(char c, FILE *f);
void fputs(char *s, FILE *f);
void fputn(int n, FILE *f);
void fputu(unsigned int n, FILE *f);
char *alloc(unsigned nbytes);
void free(char *ap);

/* code */

void main() {
  init_os();

  printf("hello world\n");
  printf("a: %d\n", alloc(200));
  printf("a: %d\n", alloc(1000));
  halt();
}

void init_os() {
  console.type = TYPE_DEV;
  console.device = DEV_CONSOLE;
  console.unit = 0;
  stdin = stdout = &console;
  printf("zen %s booting\n", VERSION);
}

int strcmp(p, q)
char *p;
char *q;
{
  int f;

  f=TRUE;
  do {
    if (*p!=*q) f=FALSE;
    q++;
  } while (*p++);
  return f;
}

FILE *fopen(name, mode) 
     char *name;
     char *mode;
{
  FILE *newfd;

  return newfd;
}

void fclose(fd)
     FILE *fd;
{
}

void fputn(n, f)
     int n;
     FILE *f;
{
  if (n<0) {
    fputc('-', f);
    n = -n;
  }
  if (n>9) fputn(n/10, f);
  fputc('0'+(n%10), f);
}

void fputu(n, f)
     unsigned int n;
     FILE *f;
{
  if (n>9) fputu(n/10, f);
  fputc('0'+(n%10), f);
}

void fputs(s, f)
     char *s;
     FILE *f;
{
  while(*s) fputc(*s++, f);
}

void printf(s, q)
     char *s;
     int q;
{
  doprintf(stdout, s, &q);
}

void fprintf(f, s, q)
     FILE *f;
     char *s;
     int q;
{
  doprintf(f, s, &q);
}

/* Oversimplistic fprintf... but enough */
void doprintf(f, s, a)
     FILE *f;
     char *s;
     int *a; /* no, it's not! */
{
  while(*s) {
    if (*s=='%') {
      switch(*++s) {
      case 's':
	fputs((char *)*(a++), f);
	break;
      case 'd':
	fputn(*(a++), f);
	break;
      case 'u':
	fputu(*(a++), f);
	break;
      case 'c':
	fputc((char)*(a++), f);
	break;
      };
    } else fputc(*s, f);
    s++;
  };
}

void fputc(c, f)
     FILE *f;
     char c;
{
  switch (f->type) {
  case TYPE_DEV:
    switch (f->device) {
    case DEV_CONSOLE:
      out(0, c);
      break;
    };
    break;
  };
}

/* The K&R alloc routines. These are stolen straight from examples in
   K&R, and are extremely clever, which is why they are used here. As
   we don't have sbrk, I've modified them to create a real list at the
   start. */

HEADER *base;
HEADER *allocp = NULL;

char *alloc(nbytes)
     unsigned nbytes;
{
  HEADER *p, *q;
  int nunits;
  
  nunits = 1+(nbytes+sizeof(HEADER)-1)/sizeof(HEADER);
  if ((q=allocp)==NULL) {
    base = (HEADER *)ramstart;
    allocp = q = base->s.ptr = base;
    base->s.size = MEMORYSIZE-ramstart;
  }
  for(p=q->s.ptr;;q=p,p=p->s.ptr) {
    if (p->s.size >= nunits) {
      if (p->s.size == nunits)
	q->s.ptr=p->s.ptr;
      else {
	p->s.size -= nunits;
	p += p->s.size;
	p->s.size=nunits;
      }
      allocp=q;
      return((char *)(p+1));
    }
    if (p==allocp)
      /*      if ((p=morecore(nunits))==NULL) */
      return(NULL); /* we don't have morecore */
  }
}

void free(ap)
     char *ap;
{
  HEADER *p, *q;
  
  p=(HEADER *)ap-1;
  for(q=allocp; !(p>q && p< q->s.ptr); q=q->s.ptr)
    if (q >= q->s.ptr && (p>q||p<q->s.ptr))
      break;
  if (p+p->s.size==q->s.ptr) {
    p->s.size += q->s.ptr->s.size;
    p->s.ptr = q->s.ptr->s.ptr;
  } else
    p->s.ptr = q->s.ptr;
  if (q+q->s.size == p) {
    q->s.size += p->s.size;
    q->s.ptr = p->s.ptr;
  } else
    q->s.ptr = p;
  allocp = q;
}
