/* Adam's calculator, mark 2 

Eventually have a line-numbers programming language not dissimilar to the
Casio language.

New ops: 
# (equals), < > ~ & | % (obv)
: (go to), of the form cond:line, e.g. a>3:12 or 1:6
multi-char variables would be nice... so would better precision,
 FP input, constants etc.
? means input a number. \ means check the keyboard and go on.
! means print a number, ' means print a character.
X means exit. $ means integer.

Guess-your-number game
 Input 0 means correct, 1 means too small, 2 means too big
1  b=0
2  t=100
3  !(a=$((t+b)/2))
4  i=?
5  i#0:11
6  i#1:9
7  b=g
8  1:3
9  t=g
10 1:3
11 X

*/

#include <stdio.h>
#include <math.h>
#include <string.h>

#define STACKSIZE 50
double valstack[STACKSIZE], var[26];
int tagstack[STACKSIZE], valsp, opstack[50], opsp;

#define NUM 'n'
#define VAR 'v'

void error(char *s) {
  printf("Err: %14s\n",s);
}

double vpush(double n, int type) {
  if (valsp==STACKSIZE) {
    error("Stack full");
    return n;
  }
  tagstack[++valsp] = type;
  printf("[push %d %c %f] ",valsp,type,n);
  return (valstack[valsp] = n);
}

int ptype;

double realvpop() {
  if (!valsp) {
    error("Stack empty");
    ptype = 0;
    return 0;
  }
  ptype = tagstack[valsp];
  printf("[pop %d %c %f] ",valsp,tagstack[valsp],valstack[valsp]);
  return valstack[valsp--];
}

double vpop() {
  double f;

  f=realvpop();
  if (ptype==VAR) {
    printf("[read var %d %f] ",ftol(f),var[ftol(f)]);
    f=var[ftol(f)];
  }
  return f;
}

int opush(int val) {
  return opstack[++opsp]=val;
}

int opop() {
  return opstack[opsp--];
}

char expr[80], *e;
char ops[]="=_SCTLENQ^e*/+-";

void do_op(int n) {
  double k;
  int i;

  printf("operate %c\n",ops[n-1]);
  switch(ops[n-1]) {
  case '^':
    k=vpop();
    vpush(pow(vpop(),k),NUM);
    break;
  case '*':
    vpush(vpop()*vpop(),NUM);
    break;
  case '/':
    k=vpop();
    vpush(vpop()/k,NUM);
    break;
  case '+':
    vpush(vpop()+vpop(),NUM);
    break;
  case '-':
    k=vpop();
    vpush(vpop()-k,NUM);
    break;
  case '_':
    vpush(0-vpop(),NUM);
    break;
  case 'S':
    vpush(sin(vpop()),NUM);
    break;
  case 'C':
    vpush(cos(vpop()),NUM);
    break;
  case 'T':
    vpush(tan(vpop()),NUM);
    break;
  case 'L':
    vpush(log10(vpop()),NUM);
    break;
  case 'N':
    vpush(log(vpop()),NUM);
    break;
  case 'E':
    vpush(exp(vpop()),NUM);
    break;
  case 'Q':
    vpush(sqrt(vpop()),NUM);
    break;
  case 'e':
    k=vpop();
    vpush(vpop()*pow(10,k),NUM);
    break;
  case '=':
    k=vpop();
    i=ftol(realvpop());
    if (ptype!=VAR) {
      printf("Not assigning to var type %d %d\n",ptype,i);
    } else {
      printf("[set var %d to %f] ",i,k);
      var[i]=k;
    }
    vpush(k,NUM);
    break;
  }
}

double expreval() {
  int m, n, initopsp;
  double scale;

  initopsp=opsp;
  m=0;
  while(*e && *e!=')') {
    if (*e>='a' && *e<='z' && *e!='e') {
      vpush((*e)-'a',VAR);
    } else if (*e=='(') {
      e++;
      vpush(expreval(),NUM);
    } else if (*e>='0' && *e<='9') {
      if (!m) {
	vpush(0,NUM);
	m=1;
      }
      if (m==2) {
	vpush(vpop()+scale*(*e-'0'),NUM);
	scale=scale/10.0;
      } else vpush(vpop()*10+(*e-'0'),NUM);
    } else if (*e=='.') {
      m=2;
      scale=0.1;
    } else {
      m=n=0;
      while(ops[n]) if(ops[n++]==*e) break;
      if(!ops[n-1]) {
	printf("bad op '%c'\n",*e);
	return 0;
      }
      if (opsp>initopsp) if (n>opstack[opsp]) do_op(opop());
      opush(n);
    }
    e++;
  }
  while(opsp>initopsp) do_op(opop());
  return vpop();
}

main() {
  valsp=opsp=0;
  strcpy(expr,"n=2*!((2+3)*4)");
  e=expr;
  printf("%s: %f\n", expr, expreval());
  strcpy(expr,"n+1");
  e=expr;
  printf("%s: %f\n", expr, expreval());
  strcpy(expr,"2^4");
  e=expr;
  printf("%s: %f\n", expr, expreval());

}

