/* --- stuff that should be in zen.h --- */

#define MEMORYTOP (48*1024)
#define VERSION "0.1"

#define NULL 0
#define EOF -1
#define FAILED 1

/* devices */
#define DEV_NONE 0
#define DEV_CONSOLE 1
#define DEV_SERIAL 2
#define DEV_PARALLEL 3
#define DEV_CONST 4

/* flags */
#define FLAG_READONLY 1
#define FLAG_EXECUTE 2
#define FLAG_ISTTY 4

struct filedescriptor {
  char type;
  int block; /* or device */
  int len; /* or unit */
  int bpos;
  int fpos;
  int inode; /* so we can track it on fclose */
  int flags;
};

typedef struct filedescriptor FILE;

struct direntry {
  int inode;
  char name[16];
};

typedef struct direntry DIRENT;

struct dirstream {
  int block;
  int bpos; /* bytes into the block */
  int fpos; /* *number* of current DIRENT */
  int len;
  DIRENT *dirent;
};

typedef struct dirstream DIR;

struct inode {
  char type;
  char refcount;
  int first; /* or device */
  int len; /* or unit */
  int flags;
};

typedef struct inode INODE;

#define IN_TYPE_NONE 0
#define IN_TYPE_FILE 1
#define IN_TYPE_DIR 2
#define IN_TYPE_DEV 3
/* IDEA: Have a TYPE_STRING, which would make sprintf a darn sight
   easier to implement. Fun, huh? */

/* The header that goes on disk blocks. */
struct blockheader {
  int next;
  int prev;
};

typedef struct blockheader BLOCK;

/* A struct that describes disk buffers. */
struct blockbuffer {
  char *buf;
  char age;
  int block;
  struct blockbuffer *next;
};
typedef struct blockbuffer BUFFER;

/* K&R.. hmm.. */
typedef int ALIGN; /* OK, this isn't a PDP-11, but what the hell */
union header {
  struct {
    union header *ptr;
    unsigned size;
  } s;
  ALIGN x;
};
typedef union header HEADER;

#define BLOCKSIZE 1024
#define NUMBLOCKS 512
#define NUMINODES 128

#define FreeBlockNum(n) ((n)/BLOCKSIZE)
#define FreeBlockIndex(n) ((n)%BLOCKSIZE)
#define InodeBlockNum(n) (FreeBlockNum(NUMBLOCKS)+1+(((n)*sizeof(INODE))/BLOCKSIZE))
#define InodeBlockIndex(n) (((n)*sizeof(INODE))%BLOCKSIZE)

#define putchar(c) fputc(c,stdout)
#define putc(c,f) fputc(c,f)
#define puts(s) (fputs(s,stdout),fputc('\n',stdout))
#define getchar() fgetc(stdin)
#define getc(f) fgetc(f)

/* The following macros are only for compatibility and should never be used */
#define gets(s) fgets(s,80,stdin)
#define exit(X) return(X);

#define isalpha(c) ((((c)>='a')&&((c)<='z'))||(((c)>='A')&&((c)<='Z')))
#define isnumber(c) (((c)>='0')&&((c)<='9'))
#define isdigit(c) isnumber(c)
#define isprint(c) ((c)>=32&&(c)<=126)
#define isatty(a) 1


